import cv2
import math
import numpy as np
from .commonfunctions import CommonFunctions

def image_resize(image, width = None, height = None, inter = cv2.INTER_AREA):
    dim = None
    (h, w) = image.shape[:2]
    if width is None and height is None:
        return image
    if width is None:
        r = height / float(h)
        dim = (int(w * r), height)
    else:
        r = width / float(w)
        dim = (width, int(h * r))
    resized = cv2.resize(image, dim, interpolation = inter)
    return resized

def find_circle_positions_with_hough_transform(img, features) -> bool:
    # Hough transform parameters
    minDist = 100
    param1 = 80
    param2 = 60
    minRadius = 5
    maxRadius = 100
    circles = cv2.HoughCircles(img, cv2.HOUGH_GRADIENT, 1, minDist, param1=param1, param2=param2, minRadius=minRadius, maxRadius=maxRadius)
    
    # Store detected hough circles in features
    features["hough_circles"] = circles
    circle_positions = []
    if circles is not None:
        for circle in circles:
            circle = circle[0]
            x = int(circle[0])
            y = int(circle[1])
            circle_positions.append((x,y))

        features["hough_circle_positions"] = circle_positions
        print(f"Found {len(circles)} hough circle{'' if len(circles) == 1 else 's'}.")
        return True
    return False


def calculate_distance(point1, point2) -> bool:
    """Calculate distance of points."""
    distance = math.sqrt((point1[0] - point2[0])**2 + (point1[1] - point2[1])**2)
    return distance

def get_color_for_contour(img, contour):
    # Gets called A LOT -> make it more efficient
    mask = np.zeros(img.shape[:2], dtype=np.uint8)
    mask = cv2.drawContours(mask, [contour], -1, 255, -1)
    col_mean = cv2.mean(img,mask)
    return (col_mean[0] + col_mean[1] + col_mean[2]) // 3

def find_center_with_contours(img_binary, img_original, output_img, features) -> bool:
    """Funky function to determine the center of the fgc.
    It tries to find a point in the image, where two overlapping contours are as close as possible to a hough transform found circle."""

    contours, _ = cv2.findContours(
        img_binary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
    )

    # Store all possible elements of the fgc in this list
    possible_fgc_elements = []
    rejected_too_few_sides_cnt = 0
    rejected_too_many_sides_cnt = 0
    rejected_too_small_cnt = 0

    for contour in contours[1:]:
        cv2.drawContours(output_img, [contour], 0, (255,0,0), 1)
        approx = cv2.approxPolyDP(contour, 0.01 * cv2.arcLength(contour, True), True)
        
        # finding center point of shape
        M = cv2.moments(contour)
        if M['m00'] != 0.0:
            x = int(M['m10']/M['m00'])
            y = int(M['m01']/M['m00'])
    
        # The number of contour sides
        contour_sides = len(approx)
        bounding_rect = cv2.minAreaRect(contour)
        (_x_br, _y_br), (width, height), _angle = bounding_rect
        bounding_rect_size = width * height

        # Only consider elements with a certain amount of contour sides and a minimum size.
        if contour_sides >= 6 and contour_sides <= 20 and bounding_rect_size >= 80:
            box = cv2.boxPoints(bounding_rect)
            box = np.int0(box)
            
            color = get_color_for_contour(img_original, contour)

            possible_fgc_elements.append(
                {"contour": contour, "x": x, "y": y, "bounding_rect": bounding_rect, "sides": contour_sides, "color": color}
            )
        else:
            if contour_sides < 6:
                rejected_too_few_sides_cnt += 1
            elif contour_sides > 20:
                rejected_too_many_sides_cnt += 1
            elif bounding_rect_size < 80:
                rejected_too_small_cnt += 1

    print(f"Rejected {rejected_too_few_sides_cnt + rejected_too_many_sides_cnt + rejected_too_small_cnt} of {len(contours[1:])} contours.")
    print(f"Too few sides:  {rejected_too_few_sides_cnt} contours.")
    print(f"Too many sides: {rejected_too_many_sides_cnt} contours.")
    print(f"Too small:      {rejected_too_small_cnt} contours.")

    # Store possible fgc elements in features
    features["possible_fgc_elements"] = possible_fgc_elements

    # Store contour pairs with their distance to each other and to the next hough transform circle in a list
    circle_pairs_with_offset_and_closeness_to_hough_circle = []
    for a, contour_dict_1 in enumerate(possible_fgc_elements):
        for b, contour_dict_2 in enumerate(possible_fgc_elements):
            # Do not consider elements themself
            if a == b:
                continue

            pair_offset = calculate_distance([contour_dict_1["x"], contour_dict_1["y"]],[contour_dict_2["x"], contour_dict_2["y"]])

            # Calculate the average center of both contours 
            average_circle_center = ((contour_dict_1["x"] + contour_dict_2["x"])/2, (contour_dict_1["y"] + contour_dict_2["y"])/2)

            # Find minimum distance to a hough circle
            min_closeness_to_hough_circle = None
            for circle_position in features["hough_circle_positions"]:
                closeness_to_hough_circle = calculate_distance(circle_position, average_circle_center)
                if min_closeness_to_hough_circle is None or closeness_to_hough_circle < min_closeness_to_hough_circle:
                    min_closeness_to_hough_circle = closeness_to_hough_circle

            # Get bounding rect sizes
            (_x_br, _y_br), (width, height), _angle = contour_dict_1["bounding_rect"]
            bounding_rect_size_a = width * height
            (_x_br, _y_br), (width, height), _angle = contour_dict_2["bounding_rect"]
            bounding_rect_size_b = width * height

            if bounding_rect_size_a > bounding_rect_size_b:
                bounding_rect_size_big = bounding_rect_size_a
                bounding_rect_size_small = bounding_rect_size_b
            else:
                bounding_rect_size_big = bounding_rect_size_b
                bounding_rect_size_small = bounding_rect_size_a

            # Store the found pair
            circle_pairs_with_offset_and_closeness_to_hough_circle.append({
                "x": int(average_circle_center[0]),
                "y": int(average_circle_center[1]),
                "bounding_rect_size_small": bounding_rect_size_small,
                "bounding_rect_size_big": bounding_rect_size_big,
                "total_sides": contour_dict_1["sides"] + contour_dict_2["sides"],
                "total_color": contour_dict_1["color"] + contour_dict_2["color"],
                "pair_offset": pair_offset,
                "closeness_to_hough_circle": min_closeness_to_hough_circle,
                "shape_a": contour_dict_1["contour"],
                "shape_b": contour_dict_2["contour"],
                "center_a": (contour_dict_1["x"], contour_dict_1["y"]),
                "center_b": (contour_dict_2["x"], contour_dict_2["y"]),
            })

    # Now we can determine the best match for our fgc center
    center_of_fgc = None
    best_score = None
    for circle_pair in circle_pairs_with_offset_and_closeness_to_hough_circle:
        # relation_big_small_score = abs(circle_pair["bounding_rect_size_small"] - (circle_pair["bounding_rect_size_big"] * 0.75))
        pair_offset_score = circle_pair["pair_offset"]
        closeness_to_hough_circle_score = circle_pair["closeness_to_hough_circle"] * 3
        side_score = circle_pair["total_sides"]
        color_score = circle_pair["total_color"] * 200
        score = pair_offset_score + closeness_to_hough_circle_score + side_score + color_score
        if best_score is None or score < best_score:
            best_score = score
            center_of_fgc = circle_pair

    # If a center is found
    if center_of_fgc is not None:
        x_a, y_a, w_a, h_a = cv2.boundingRect(center_of_fgc["shape_a"])
        x_b, y_b, w_b, h_b = cv2.boundingRect(center_of_fgc["shape_b"])
        if w_a * h_a < w_b * h_b:
            true_center = center_of_fgc["center_a"]
            center_shape = center_of_fgc["shape_a"]
            orientation_shape = center_of_fgc["shape_b"]
        else:
            true_center = center_of_fgc["center_b"]
            center_shape = center_of_fgc["shape_b"]
            orientation_shape = center_of_fgc["shape_a"]
        
        # Store the center to the features
        features["center_circle"] = center_shape
        features["orientation_ring"] = orientation_shape
        features["center_coordinates"] = true_center
        return True
    return False


def find_orientation_dot(features) -> None:
    """Try to find the orientation dot by determing the contour with the minimum distance to the fgc center."""

    for possible_fgc_element in features["possible_fgc_elements"]:
        possible_fgc_element["distance_to_center"] = calculate_distance((possible_fgc_element["x"], possible_fgc_element["y"]), features["center_coordinates"])
    
    sorted_possible_fgc_elements = sorted(features["possible_fgc_elements"], key=lambda elem: elem["distance_to_center"])

    features["possible_fgc_elements"] = sorted_possible_fgc_elements

    possible_orientation_dot = features["possible_fgc_elements"][2]
    _x,_y,w,h = cv2.boundingRect(possible_orientation_dot["contour"])
    _x2,_y2,w2,h2 = cv2.boundingRect(features["center_circle"])
    possible_orientation_dot_area = w * h
    center_circle_area = w2 * h2
    while (possible_orientation_dot_area >= center_circle_area):
        features["possible_fgc_elements"].pop(2)
        if len(features["possible_fgc_elements"]) > 2:
            possible_orientation_dot = features["possible_fgc_elements"][2]
            _x,_y,w,h = cv2.boundingRect(possible_orientation_dot["contour"])
            possible_orientation_dot_area = w * h
        else:
            break
    features["orientation_dot"] = possible_orientation_dot


def sanitize_data(features) -> None:
    """Try to get rid of all contours outside of the fgc by setting a max jump distance between contour distances to the center."""

    # Add furthest_distance_to_center entry to all contours
    for element in features["possible_fgc_elements"]:
        furthest_distance_to_center = 0
        for point in element["contour"]:

            point_distance = calculate_distance(features["center_coordinates"], point[0]) 
            if point_distance > furthest_distance_to_center:
                furthest_distance_to_center = point_distance
        element["furthest_distance_to_center"] = furthest_distance_to_center

    # Sort contours by furthest_distance_to_center
    features["possible_fgc_elements"] = sorted(features["possible_fgc_elements"], key=lambda elem: elem["furthest_distance_to_center"])

    # Remove all contours which are too far from the previous contour
    max_jump_distance = features["orientation_dot"]["distance_to_center"] * 0.5
    last_distance = features["possible_fgc_elements"][1]["furthest_distance_to_center"]
    index = 2
    while index < len(features["possible_fgc_elements"]):
        element = features["possible_fgc_elements"][index]
        if element["furthest_distance_to_center"] - last_distance > max_jump_distance:
            features["possible_fgc_elements"].pop(index)
        else:
            last_distance = element["furthest_distance_to_center"]
            element["index"] = index
            index += 1

def get_all_angles(features) -> None:
    """Get angles of all shapes regarding to the orientation dot."""

    vector_from_center_to_orientation_point = [
        features["orientation_dot"]["x"] - features["center_coordinates"][0],
        features["orientation_dot"]["y"] - features["center_coordinates"][1],
    ]

    for element in features["possible_fgc_elements"]:
        vector_from_center_to_element = [
            element["x"] - features["center_coordinates"][0],
            element["y"] - features["center_coordinates"][1],
        ]

        unit_vector_1 = vector_from_center_to_orientation_point / np.linalg.norm(vector_from_center_to_orientation_point)
        unit_vector_2 = vector_from_center_to_element / np.linalg.norm(vector_from_center_to_element)
        dot_product = np.dot(unit_vector_1, unit_vector_2)
        clipped_dot_product = np.clip(dot_product, -1.0, 1.0)
        rad = np.arccos(clipped_dot_product)
        if vector_from_center_to_orientation_point[0] * vector_from_center_to_element[1] - vector_from_center_to_orientation_point[1] * vector_from_center_to_element[0] < 0:
            rad = -rad
        deg = 0
        if not math.isnan(rad):
            deg = math.degrees(rad)
            deg = math.floor(deg+0.5)
            if deg < 0:
                deg = 180 + (180 - abs(deg))
            if deg >= 359:
                deg = 0

        element["angle"] = deg


def divide_elements_into_rings_by_angle_and_distance(features):
    current_ring = 0
    current_distance = 0
    rings = []
    angle_sorted_rings = []

    # Divide elements into rings by furthest_distance_to_center
    for element in features["possible_fgc_elements"]:

        # Increase ring if there is a jump in distance to center
        if abs(current_distance - element["furthest_distance_to_center"]) >= features["orientation_dot"]["distance_to_center"] * 0.2:
            current_ring += 1
        
        # Append new ring to rings if necessary and append current contour element
        if current_ring >= len(rings):
            rings.append([])
        
        # Current ring append contour
        rings[current_ring - 1].append(element)

        # Set current distance to own maximum distance
        current_distance = element["furthest_distance_to_center"]

    # Get number of dots for each contour
    for ring in rings:
        if len(ring) < 1: continue
        sorted_ring = sorted(ring, key=lambda elem: elem["angle"])
        angle_sorted_rings.append(sorted_ring)

    features["rings"] = angle_sorted_rings


def rotate_vector(v,deg):
    v = np.array(v)
    assert len(v)==2
    
    phi = np.deg2rad(deg)
    s = np.sin(phi)
    c = np.cos(phi)
    M = np.array([[c,-s],[s, c]])

    return M.dot(v)


def get_data_from_rings(features):
    currently_zero = True
    data_rings = []
    data = []
    
    features["target_positions"] = []
    features["data_rings"] = []
    features["data"] = []

    print(f"Getting data of { len(features['rings']) } rings...")

    for ring_id, ring in enumerate(features["rings"]):

        # Skip first two "rings" because they are the center circle and the orientation ring+dot
        if ring_id < 2:
            continue

        ring_id = ring_id - 1

        data_ring = []
        currently_zero = True   # Every ring begins with a zero
        vector_from_center_to_orientation_point = [
            (features["orientation_dot"]["x"] - features["center_coordinates"][0]),
            (features["orientation_dot"]["y"] - features["center_coordinates"][1]),
        ]
        vector_from_center_to_dot = [
            vector_from_center_to_orientation_point[0] / 2 + ((vector_from_center_to_orientation_point[0] / 2) * (ring_id + 1)),
            vector_from_center_to_orientation_point[1] / 2 + ((vector_from_center_to_orientation_point[1] / 2) * (ring_id + 1)),
        ]

        degrees_per_bit = CommonFunctions.get_degrees_per_bit(ring_id)
        numBits = int( 360 / degrees_per_bit )
        
        current_contour_angle = None

        for pos in range(numBits):
            rotated_vector = vector_from_center_to_dot
            if pos > 0:
                rotated_vector = rotate_vector( vector_from_center_to_dot, degrees_per_bit * pos )
            # Get position and check in which contour it lands in
            target_position = [
                int(features["center_coordinates"][0] + rotated_vector[0]), 
                int(features["center_coordinates"][1] + rotated_vector[1])
            ]

            # Check all contours if they contain the calculated position
            found_contour = False
            closest_contour = None
            total_closest_point_distance = None
            for element in ring:
                is_point_in_contour = cv2.pointPolygonTest(element["contour"], (target_position[0], target_position[1]), False)
                if is_point_in_contour == 1:
                    # If target position lies within this contour, this is the one we are looking for
                    if pos > 0:
                        if current_contour_angle is not None and current_contour_angle != element["angle"]:
                            currently_zero = not currently_zero
                        if currently_zero:
                            data_ring.append(0)
                            data.append(0)
                        else:
                            data_ring.append(1)
                            data.append(1)

                    found_contour = True
                    current_contour_angle = element["angle"]
                    break
                else:
                    # If target position lies outside contour, keep track of the distance to this contour
                    closest_point_distance_in_contour = None
                    for point in element["contour"]:
                        distance = calculate_distance(target_position, point[0])
                        if closest_point_distance_in_contour is None or distance < closest_point_distance_in_contour:
                            closest_point_distance_in_contour = distance

                    if total_closest_point_distance is None or closest_point_distance_in_contour < total_closest_point_distance:
                        total_closest_point_distance = closest_point_distance_in_contour
                        closest_contour = element

            if not found_contour:
                # If contour was not found yet, check the closest distance to a contour and decide if it should have been inside
                if closest_contour is None or total_closest_point_distance > features["orientation_dot"]["distance_to_center"] * 0.2:
                    break
                else:
                    if pos > 0:
                        if current_contour_angle is not None and current_contour_angle != closest_contour["angle"]:
                            currently_zero = not currently_zero
                        if currently_zero:
                            data_ring.append(0)
                            data.append(0)
                        else:
                            data_ring.append(1)
                            data.append(1)

                    current_contour_angle = closest_contour["angle"]

            features["target_positions"].append(target_position)

        data_rings.append(data_ring)

    features["data_rings"] = data_rings
    features["data"] = data


import cv2
import matplotlib.pyplot as plt


def set_brightness_contrast(img, brightness=255, contrast=127):
    brightness = int((brightness - 0) * (255 - (-255)) / (510 - 0) + (-255))
    contrast = int((contrast - 0) * (127 - (-127)) / (254 - 0) + (-127))
    if brightness != 0:
        if brightness > 0:
            shadow = brightness
            max = 255
        else:
            shadow = 0
            max = 255 + brightness
 
        al_pha = (max - shadow) / 255
        ga_mma = shadow
        cal = cv2.addWeighted(img, al_pha, img, 0, ga_mma)
    else:
        cal = img
 
    if contrast != 0:
        Alpha = float(131 * (contrast + 127)) / (127 * (131 - contrast))
        Gamma = 127 * (1 - Alpha)
        cal = cv2.addWeighted(cal, Alpha, cal, 0, Gamma)

    return cal

def show_image(unique_name, image, title=None):
    plt_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    plt.figure(unique_name)
    if title != None and type(title) is str:
        plt.title(title)
    plt.imshow(plt_image)
    plt.show(block=False)




#!/usr/bin/env python3
"""Fancy Galaxy Code (FGC) entrypoint.

Before execution you have to install the requirements by executing:
pip install -r requirements.txt

Execution:
python3 test_creator.py 'Content of fgc' outputfile.svg
"""

import sys
from fgc_tools import FGCCreator


def main() -> None:
    """Entrypoint for creating an FGC."""
    data: str = ""
    output_file_name: str = "fgc.svg"
    color_inner: str = "#009060"
    color_outer: str = "#006090"
    background_color: str = "#ffffff"
    write_data_as_text = True

    # Store arguments into variables
    argument_count: int = len(sys.argv)
    if argument_count > 1:
        data = sys.argv[1]
        data = data.replace("\\n", "\n")
    if argument_count > 2:
        output_file_name = sys.argv[2]
    if argument_count > 4:
        color_inner = sys.argv[3]
        color_outer = sys.argv[4]
    if argument_count > 5:
        background_color = sys.argv[5]
    if argument_count > 6:
        write_data_as_text = (sys.argv[6].lower() == 'true')
    

    print("Fancy Galaxy Code (FGC) creator")
    FGCCreator.create_fgc(
        color_inner=color_inner, 
        color_outer=color_outer,
        data=data, 
        output_file=output_file_name,
        color_background=background_color,
        write_data_as_text=write_data_as_text
    )


if __name__ == "__main__":
    main()



import cv2
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
import scipy
from cvfunctions import *
from commonfunctions import CommonFunctions
from commonconstants import CommonConstants




def rotate_vector(v,deg):
    v = np.array(v)
    assert len(v)==2
    
    phi = np.deg2rad(deg)
    s = np.sin(phi)
    c = np.cos(phi)
    M = np.array([[c,-s],[s, c]])

    return M.dot(v)


def check_contour_for_color(img, contour, color, color_precision = 255/10):
    """Will check img in contour region for that average color
    Will return value of average color if color argument is None"""

    mask = np.zeros(img.shape[:2], dtype=np.uint8)
    mask = cv2.drawContours(mask, [contour], 0, 255, -1)
    
    col_mean = cv2.mean(img,mask)
    if type(color) not in [list, tuple]:
        col_mean = col_mean[0]
    else:
        col_mean = col_mean[:len(color)]
    
    if color is None:
        return col_mean
    else:
        diff = np.abs(color - col_mean)
        return diff <= color_precision


def check_contour_for_spherical(contour):
    ellipse = cv2.fitEllipse(contour) # ( (position), (FULL axes), rotation)
    axes = ellipse[1] # these are NOT semi-axes!!!
    a = max(axes) / 2
    b = min(axes) / 2
    ratio_axes = b/a # ratio between the axes: 1=circle
    
    e_sq = 1.0 - ratio_axes**2  # eccentricity squared
    arc_ellipse = 4*a*scipy.special.ellipe(e_sq)  # circumference formula
    arc_cont = cv2.arcLength(contour, True)
    arcs = [arc_ellipse, arc_cont]
    ratio_arc = min(arcs) / max(arcs)
    
    area_ellipse = min(axes) * max(axes) * np.pi / 4
    area_cont = cv2.contourArea(contour)
    areas = [area_ellipse, area_cont]
    ratio_area = min(areas) / max(areas)

    ratio_geom_mean = np.power(ratio_axes * ratio_area * ratio_arc, 1/3) # geometric mean between all ratios
    
    return ratio_geom_mean, ratio_axes, ratio_area, ratio_arc


def find_dot_orientation(img_bin, center, radius):
    assert len(img_bin.shape)==2
    
    img = img_bin.copy()
    img_negative = img.copy()
    # show_image(1, img)
    
    center = np.array(center)
    c1 = np.array(np.round(center), dtype=np.int)
    # c1 = np.flip(c1)
    
    r_mid = int( np.round(radius * 6/4) )
    
    f_radi = [1.125, 1.875]
    r_min = int( np.round(radius * f_radi[0]) )
    r_max = int( np.round(radius * f_radi[1]) )
    
    img_t = np.zeros(img.shape[:2], dtype=np.uint8) # template for first ring
    # cv.circle(	img, center, radius, color[, thickness
    img_t = cv2.circle(img_t, c1, r_max, [255], -1)
    img_t = cv2.circle(img_t, c1, r_min, 0, -1)
    
    img = cv2.bitwise_and(img_negative, img_t)
    show_image(1, img)
    
    img_1 = np.zeros(img.shape, dtype=np.uint8)
    img_1 = cv2.circle(img_1, c1, r_mid, 255, 5)
    
    img_2 = cv2.bitwise_xor(img, img_1, mask=img_1)
    img_2 = cv2.dilate(img_2, np.ones((3,3))*255)
    show_image(2, img_2)
    
    contours, _ = cv2.findContours(img_2, cv2.RETR_EXTERNAL , cv2.CHAIN_APPROX_SIMPLE)
    assert len(contours)==2
    
    pts = []
    for cont in contours:
        M = cv2.moments(cont)
        cx = M['m10']/M['m00']
        cy = M['m01']/M['m00']
        cxy = (np.array([cx,cy]) - c1)*2 + c1
        cxy = np.array(cxy, dtype=np.int)
        pts.append( cxy )

    pts.append(c1)
    con_3 = [np.array([val]) for val in pts]
    con_3 = np.array(con_3)
    
    img_t = np.zeros(img.shape[:2], dtype=np.uint8) # template for first ring
    img_t = cv2.drawContours(img_t, [con_3], 0, 255, -1)
    img_3 = cv2.bitwise_and(img, img_t, mask=img_t)
    show_image(3, img_3)
    
    contours, _ = cv2.findContours(img_3, cv2.RETR_EXTERNAL , cv2.CHAIN_APPROX_SIMPLE)
    assert len(contours)==1
    
    cont = contours[0]
    cxy = cv2.fitEllipse(cont)[0]
    # cxy = np.array(cxy, dtype=np.int)
    v_dot = cxy-center
    # r_dot = np.sqrt( np.sum( np.square(c_dot) ) )
    # angle = np.arctan2(c_dot[1],c_dot[0]) * 180/np.pi
    
    return v_dot


def find_dot_orientation_2(img_bin, center, vector_0):
    assert len(img_bin.shape)==2
    img_debug = img_bin.copy()
    r_max = int(img_bin.shape[0]/2)
    r_0 = 60
    r_0_real = np.sqrt( np.sum( np.square(vector_0) ) )
    v_center = np.array(center)
    
    vU = vector_0 / r_0_real # unit vector
    Rscale = r_0_real/r_0
    
    ret_list = [vector_0]
    step = 30
    
    timeout = 20
    r = r_0 + step
    n = 1
    while r < r_max and timeout > 0:
        
        pt1 = np.array(v_center + vU * (r-step/2) * Rscale, dtype=int)
        pt2 = np.array(v_center + vU * (r+step/2) * Rscale, dtype=int)
        print(pt1 - pt2, pt1, pt2)
        
        # show_image(1, img_bin,"binary")
        img_temp = np.zeros(img_bin.shape, dtype=np.uint8)
        img_temp = cv2.line(img_temp, pt1, pt2, 255, 3)
        # show_image(2, img_temp, "temp line")
        img_temp = cv2.bitwise_and(img_bin, img_temp, mask=img_temp)
        # show_image(3, img_temp, "result of detection")
        # print("r:", r)
        
        
        contours, _ = cv2.findContours(img_temp, cv2.RETR_EXTERNAL , cv2.CHAIN_APPROX_SIMPLE)
        contours = list(contours)
        contours.sort(key= lambda x: cv2.contourArea(x), reverse=True) # biggest area
        cont = contours[0]
        
        M = cv2.moments(cont)
        assert M['m00'] > 0
        cx = M['m10']/M['m00']
        cy = M['m01']/M['m00']
        v_ci = np.array([cx,cy])
        
        ret_list.append(v_ci-v_center)
        
        n += 1
        r += step
        timeout -= 1
    
    return ret_list


def main() -> None:
    DEBUG_FLAG = False

    with open("_used_versions.txt","w") as f:
        f.write("cv2: {}\n".format(cv2.__version__) )
        f.write("numpy: {}\n".format(np.__version__) )
        f.write("matplotlib: {}\n".format(mpl.__version__) )
        f.write("seaborn: {}\n".format(sns.__version__) )
        f.write("scipy: {}\n".format(scipy.__version__) )

    test_images = [
        { "img": 'test_images/1.jpg', "content": "Level1" },
        { "img": 'test_images/2.jpg', "content": "Level2 which is harder" },
        { "img": 'test_images/3.jpg', "content": "Level3 which is super hard" },
        { "img": 'test_images/4.jpg', "content": "Level1" },
        { "img": 'test_images/5.jpg', "content": "Level2 which is harder" },
    ]

    path_test_image = test_images[0]["img"]

    dict_imgs={}

    img = cv2.imread(path_test_image)
    dict_imgs["high_contrast"] = img.copy()
    dict_imgs["high_contrast"] = set_brightness_contrast(dict_imgs["high_contrast"], 190, 178)

    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    dict_imgs["20 gray"] = img.copy()

    _, img = cv2.threshold(img, 127, 255, cv2.THRESH_BINARY)
    dict_imgs["30 binary"] = img.copy()

    show_image(1, dict_imgs["30 binary"])

    img = dict_imgs["30 binary"]
    contours, _ = cv2.findContours(img, cv2.RETR_LIST , cv2.CHAIN_APPROX_SIMPLE)
    contours_binary = contours

    cont_2 = {}

    for i, contour in enumerate(contours):
        cont_2[i] = {"isDark":False, "sphereRating":0, "center":(0,0), "area":0}
        
        M = cv2.moments(contour)
        if M['m00'] > 0 :
            cx = int( M['m10']/M['m00'] )
            cy = int( M['m01']/M['m00'] )
            
            cont_2[i]["area"] = M['m00']
            cont_2[i]["center"] = ( int(cx), int(cy) )
        
        cont_2[i]["isDark"] = check_contour_for_color(img, contour, 0, 127)
        if not cont_2[i]["isDark"]: continue
        
        ratios = check_contour_for_spherical(contour)
        cont_2[i]["sphereRating"] = ratios[0]



    list_area = [entry["area"] for k,entry in cont_2.items() if (entry["isDark"]==True) and (entry["area"] > 0)]
    list_area.sort()

    n_bins = (max(list_area)-min(list_area))/100 # we want a bin width of abt 75
    n_bins = int(n_bins)
    if DEBUG_FLAG: plt.hist(list_area,n_bins)

    n,bins = np.histogram(list_area,n_bins)
    i_max = np.argmax(n)
    area_min = bins[i_max]
    area_max = bins[i_max+1]
    list_area = [a for a in list_area if a> area_min and a < area_max]

    area_median = np.mean(list_area)
    if DEBUG_FLAG: 
        plt.axvline(x=area_median,color="red", linestyle="dashed")
        plt.show()

    print("Average most common circle area: {:.2f}".format(area_median) )


    del DEBUG_FLAG

    img = dict_imgs["20 gray"]
    img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)

    list_dot_area = []
    for i,contour in enumerate(contours):
        col = [0,255,0]
        thick = 2
        txt = "{:>2} ".format(i)
        entry = cont_2[i]
        if not entry["isDark"]: 
            col = [255,127,0]
            thick = 1
        else:
            if entry["area"] >= 10:
                txt += "{:6.2f} {:6.0f}".format(entry["sphereRating"],entry["area"])
                if entry["sphereRating"] < 0.9: col = [0,0,255]

        img = cv2.drawContours(img, [contour], 0, col, thick)
        img = cv2.putText(img, txt, (10,5+12*(i+1)), cv2.FONT_HERSHEY_PLAIN, 1, col)
        img = cv2.putText(img, str(i), entry["center"], cv2.FONT_HERSHEY_PLAIN , 1, col)
        

    show_image("First", img)

    a = np.array([[1, 2]])
    b = np.array([[4, 1]])
    c = np.outer(a, b)

    v1 = np.ones(len(cont_2))
    for i, entry in cont_2.items():
        if entry["isDark"]==True and (entry["sphereRating"] > 0.9):
            v1[i] = entry["area"]
    v2 = np.array([1/v for v in v1])

    z1 = np.outer(v1,v2)
    z1 = z1/16
    z1 = np.log(z1)
    z1 = np.abs(z1)
    # print(np.min(z1))
    z1_flat = z1.flatten()

    z2 = [(i,v) for i,v in enumerate(z1_flat)]
    z2.sort(key = lambda val: val[1])
    z2 = [val for val in z2 if val[1] < 1]
    if len(z2)>10: z2 = z2[:10]

    points = {}
    for entry in z2:
        pos = entry[0]
        idx = np.unravel_index(pos, z1.shape)[0]
        if idx not in points: points[idx]=0
        
        points[idx] += 1
        pass

    point_ranking = []
    for key,val in points.items():
        point_ranking.append((key,val))
        pass

    point_ranking.sort(key = lambda val: val[1], reverse = True)
    center_idx = point_ranking[0][0]
    print("Center circle is at idx: {}".format(center_idx))

    img_negative = 255 - dict_imgs["30 binary"]
    img = img_negative.copy()

    cont_center = contours_binary[center_idx]
    ellipse = cv2.fitEllipse(cont_center)
    pos_center = np.array(ellipse[0])
    axes = np.array(ellipse[1])
    radius = np.mean(axes)/2


    kernel = np.ones((5,5), dtype=np.uint8) * 255
    iterations = int( radius/4 )

    img = cv2.dilate(img, kernel, iterations=iterations)
    show_image(1, img)

    img = cv2.erode(img, kernel, iterations=iterations-1)
    show_image("Some name", img)

    contours, _ = cv2.findContours(img, cv2.RETR_LIST , cv2.CHAIN_APPROX_SIMPLE)
    cont_new = None

    check_contours = []
    for cont in contours:
        if(check_contour_for_color(img, cont, 255, 10)):
            check_contours.append(cont)
            
            #np.array(pos_center, dtype=np.int)
            if cv2.pointPolygonTest(cont,pos_center, False) > 0:
                cont_new = cont
                break


    img_t = np.zeros(img.shape[:2], dtype=np.uint8)
    img_t = cv2.drawContours(img_t, [cont_new], 0, 255, -1)
    img_t = cv2.dilate(img_t, kernel, iterations=iterations)

    img = cv2.bitwise_and(img_negative, img_t)

    boundRect = cv2.boundingRect(cont_new)
    pt1 = np.array( (int(boundRect[0]), int(boundRect[1])) )
    pt2 = np.array( (int(boundRect[0]+boundRect[2]), int(boundRect[1]+boundRect[3])) )

    l_dist = np.array([pt1-pos_center, pt2-pos_center]).flatten()
    r = int( np.ceil( np.max( np.abs(l_dist) ) + radius/4 ) )
    vr = np.array([r,r])
    pt1 = np.array(pos_center - vr, dtype=np.int)
    pt2 = np.array(pos_center + vr, dtype=np.int)

    #get Region of Interest
    img = img_negative[ pt1[1]:pt2[1],pt1[0]:pt2[0] ].copy()

    # Scale to standard size
    scale_f = 40/radius
    if scale_f > 1.01:
        img = cv2.resize(img, None, fx=scale_f, fy=scale_f, interpolation=cv2.INTER_CUBIC )
    if scale_f < 0.99:
        img = cv2.resize(img, None, fx=scale_f, fy=scale_f, interpolation=cv2.INTER_AREA )

    # threshold scaled image
    _, img = cv2.threshold(img, 127, 255, cv2.THRESH_BINARY)

    # store work image
    dict_imgs["40 work"] = img.copy()
    cv2.imwrite("out_temp 3.png", img)

    show_image(3, img)


    #%%
    #plt.close("all")

    # img = dict_imgs["20 gray"]
    img = dict_imgs["40 work"]
    contours, _ = cv2.findContours(img, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

    img = cv2.cvtColor( img, cv2.COLOR_GRAY2BGR )
    img_ov = np.zeros(img.shape, dtype=np.uint8)

    list_contours = []
    a = input()
    for i,cont in enumerate(contours):
        ratios = check_contour_for_spherical(cont)
        ratio = ratios[0]
        isShpere = ratio > 0.95
        area = cv2.contourArea(cont)
        
        list_contours.append( {"idx":i,
                            "cont": cont,
                            "ratio": ratio,
                            "isShpere":isShpere,
                            "area": area } )
        pass

    list_sphere = []
    for entry in list_contours:
        if entry["isShpere"]:
            list_sphere.append(entry)
            pass
        pass

    list_sphere.sort(key = lambda val: val["area"], reverse = True)
    dot_center = list_sphere.pop(0)


    img_ov = cv2.drawContours(img_ov, [dot_center["cont"]], 0, [255,127,0], 5)

    list_areas = []
    for entry in list_sphere:
        img_ov = cv2.drawContours(img_ov, [entry["cont"]], 0, [0,255,0], 5)
        list_areas.append(entry["area"])
        pass

    sphere_area = np.mean(list_areas)


    img_out = cv2.addWeighted(img, 0.5, img_ov, 0.5, 0)
    show_image(1, img_out)



    #%%
    #plt.close("all")

    # img = dict_imgs["20 gray"]
    img = cv2.cvtColor( dict_imgs["40 work"], cv2.COLOR_GRAY2BGR )
    img_ov = np.zeros(img.shape, dtype=np.uint8)

    cont_center = dot_center["cont"]
    ellipse = cv2.fitEllipse(cont_center)
    pos_center = np.array(ellipse[0])
    axes = np.array(ellipse[1])/2
    angle = ellipse[2]

    radius = np.mean(axes)
    c1 = np.array( np.round(pos_center), dtype=np.int )
    ax = np.array( np.round(axes), dtype=np.int )


    vector_dot = find_dot_orientation(dict_imgs["40 work"], pos_center, radius)

    ret = find_dot_orientation_2(dict_imgs["40 work"],pos_center,vector_dot)

    v_dots = np.array(ret)
    radii = np.sqrt( np.sum( np.square(v_dots), 1 ) )
    angles = np.arctan2(v_dots[:,1], v_dots[:,0])
    # np.rad2deg(angles)
    phi_0 = np.mean([angles[0], angles[-1]])
    # np.rad2deg(phi_0)


    # raise Exception()

    for i,vR in enumerate(radii):
        r = int(np.round(vR) )
        img_ov = cv2.circle(img_ov, c1, int(r), [0,0,255], 1)
        if i>0:
            dpb = CommonFunctions.get_degrees_per_bit(i)
            for pos in range(int(360/dpb)):
                vi = rotate_vector(v_dots[i], dpb*pos)
                ci = np.array( np.round(pos_center + vi), dtype=np.int )
                img_ov = cv2.circle(img_ov, ci, 10, [255,0,0], -1)

    img_out = cv2.addWeighted(img, 0.5, img_ov, 0.5, 0)
    show_image(1, img_out)
    cv2.imwrite("out_temp 4.png", img_out)

    img_bin = dict_imgs["40 work"].copy()
    img = cv2.cvtColor( dict_imgs["40 work"], cv2.COLOR_GRAY2BGR )

    dict_rings = {}

    idx=0

    bit_string = ""

    for i,vDot in enumerate(v_dots):
        if i==0: continue
        
        
        dpb = CommonFunctions.get_degrees_per_bit(i)
        numBits = int(360/dpb)-1
        dict_rings[i] = []
        
        bitVal = False
        
        for pos in range(numBits):
            v_valid = rotate_vector(vDot, dpb * (pos + 1) )
            v_valid_c = np.array( np.round(pos_center + v_valid), dtype=np.int )
            bit_valid = img_bin[v_valid_c[1], v_valid_c[0] ] > 127 # check if the next dot exists for the bit inbetween to count
            if not bit_valid: break
            
            v_info = rotate_vector(vDot, dpb * (pos + 1/2) )
            v_info_c = np.array( np.round(pos_center + v_info), dtype=np.int )
            bit_info = img_bin[v_info_c[1], v_info_c[0] ] > 127 # check if bit is set
            dict_rings[i].append(1 if bit_info else 0)
            bitVal = bitVal if bit_info else not bitVal
            bit_string += str(1 if bitVal else 0)
            
            col = [63,200,0] if bit_info else [0,0,255]
            marker = cv2.MARKER_DIAMOND if bit_info else cv2.MARKER_TILTED_CROSS
            img = cv2.drawMarker(img, v_info_c, col, marker,5,3)
            
        

    for ring in dict_rings:
        print("ring",ring,dict_rings[ring])
    print(bit_string)

    show_image(2, img)
    cv2.imwrite("out_temp 5.png", img)


if __name__ == '__main__':
    main()



import numpy as np
import cv2
import time
from .cvfunctions import *
from .featurehandler import *
from .libs.hamming import *


# Currently used fgc reader

class FGCReader():

    def read_image(image_path=None, image_file=None) -> str:
        np.seterr(invalid='ignore')

        # Run operations on img and draw on output_img
        if image_path:
            img = cv2.imread(image_path)
        elif image_file:
            image_path = "Memory"
            img = cv2.imdecode(np.fromstring(image_file, np.uint8), 1)
        else:
            return ("", [], 0, "-", None, None)

        # Get image dimensions
        height = img.shape[0]
        width = img.shape[1]
        print("Width:  ", width)
        print("Height: ", height)

        # Resize images if they are too large
        max_w_h = 2500
        if width > height and width > max_w_h:
            img = image_resize(img, width=max_w_h)
        elif height > max_w_h:
            img = image_resize(img, height=max_w_h)

        # Get image dimensions after resize
        height = img.shape[0]
        width = img.shape[1]

        # Features is used to store a lot of useful information 
        features = {
            "center_coordinates": None,
            "center_circle": None,
            "orientation_ring": None,
            "orientation_dot": None,
            "hough_circle_positions": None,
            "hough_circles": None
        }

        # Measure time of calculations for optimization purposes (since it will be re-written in C++ for mobile devices later)
        start_time = time.time()

        # Calculate some alternative representations of the input image
        output_img = img.copy()
        gray_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        blurred_gray_img = cv2.medianBlur(gray_img, 3)
        _, binary_img = cv2.threshold(gray_img, 150, 255, cv2.THRESH_BINARY)

        if find_circle_positions_with_hough_transform(blurred_gray_img, features):
            if find_center_with_contours(binary_img, img, output_img, features):

                # Now that we've found the center circle of the fgc, do some more examination of the data and contours
                find_orientation_dot(features)
                sanitize_data(features)
                get_all_angles(features)
                divide_elements_into_rings_by_angle_and_distance(features)
                get_data_from_rings(features)

                # Print some info on the output image
                cv2.circle(output_img, (features["center_coordinates"][0], features["center_coordinates"][1]), 4, (255,255,255), 2)

                for target_position in features["target_positions"]:
                    cv2.drawMarker(output_img, (target_position[0], target_position[1]), (0,255,0), cv2.MARKER_TILTED_CROSS, 5, 1)

                for ring_id, ring in enumerate(features["rings"]):
                    for element_id, element in enumerate(ring):
                        # cv2.circle(output_img, (element["x"], element["y"]), 4, (255,255,0), 1)

                        outline_color = (255,255,0)
                        if ring_id % 2:
                            outline_color = (0,255,255)
                        cv2.drawContours(output_img, [element["contour"]], 0, outline_color, 1)

                        # cv2.putText(
                        #     output_img, 
                        #     str(ring_id) + ":" + str(element_id), 
                        #     (element["x"], element["y"] - 15), 
                        #     cv2.FONT_HERSHEY_SIMPLEX, 
                        #     0.5, (0,190,0), 1, cv2.LINE_AA
                        # )
                        # cv2.putText(
                        #     output_img, 
                        #     str(int(element["angle"])) + " deg", 
                        #     (element["x"], element["y"]), 
                        #     cv2.FONT_HERSHEY_SIMPLEX, 
                        #     0.5, (190,0,0), 1, cv2.LINE_AA
                        # )
                        # cv2.putText(
                        #     output_img, 
                        #     str(int(element["furthest_distance_to_center"])), 
                        #     (element["x"], element["y"] + 15), 
                        #     cv2.FONT_HERSHEY_SIMPLEX, 
                        #     0.5, (0,0,190), 1, cv2.LINE_AA
                        # )

                # Draw outline of center and orientation ring
                cv2.drawContours(output_img, [features["center_circle"]], 0, (0, 0, 255), 1)
                cv2.drawContours(output_img, [features["orientation_ring"]], 0, (0, 0, 255), 1)
                cv2.drawContours(output_img, [features["orientation_dot"]["contour"]], 0, (0, 255, 0), 1)

                print("Processed FGC successfully.")
            else:
                print("Could not find center of circle.")
        else:
            print("Could not find FGC at all.")
        
        read_time = (time.time() - start_time)

        try:
            raw_binary_string = ''.join([str(ch) for ch in features["data"]])
            # show_image("Result of " + image_path, output_img)

            all_data_decoded = hamming_decode(features["data"])
            version = all_data_decoded[:4]
            text = all_data_decoded[4:]
            binary_string = ''.join([str(ch) for ch in text])

            byte_seq = bytes(int(binary_string[i:i+8], 2) for i in range(0, len(binary_string), 8))
            str_data = byte_seq.decode('utf-8', 'ignore')

            # Strip last 0 away
            if str_data[-1] == 0:
                str_data = str_data[:-1]

            return (str_data, version, read_time, raw_binary_string, output_img, binary_img)
        except Exception as ex:
            print(ex)
            return ("", [], read_time, "-", output_img, binary_img)



from bitarray import bitarray


def print_bitarray(bit_array, end="\n"):
    current_bit = 0
    bits_per_line = 60
    for bit in bit_array:
        current_bit += 1
        print("%i" % (1 if bit else 0), end="")
        if current_bit % 60 == 0:
            print()
        elif current_bit % 4 == 0:
            print("", end=" ")
    print("", end=end)


def string_to_bitarray(string):
    bit_array: bitarray = bitarray()
    for c in string:
        bit_array.extend(byte_to_bitarray(c, 8))
    return bit_array


def byte_to_bitarray(data, lenght_in_bits):
    bit_array: bitarray = bitarray()
    if type(data) == str:
        data = ord(data)
    for i in range(lenght_in_bits):
        bit_array.append((bool)((data >> i) & 1))
    return bit_array[::-1]



from .commonconstants import CommonConstants

class CommonFunctions():
    
    @staticmethod
    def get_degrees_per_bit(ring_number): 
        if ring_number <= len(CommonConstants.DEGREES_PER_BIT):
            return CommonConstants.DEGREES_PER_BIT[ring_number-1]
        return 3



from fgc_tools import FGCReader
import matplotlib.pyplot as plt
from fgc_tools.cvfunctions import *


def main():
    print("FGC Reader started")

    test_images = [
        { "img": 'static/test_images/1.jpg', "content": "Level1" },
        { "img": 'static/test_images/2.jpg', "content": "Level2 which is harder" },
        { "img": 'static/test_images/3.jpg', "content": "Level3 which is super hard" },
        { "img": 'static/test_images/4.jpg', "content": "Meister der Galaxie." },
        { "img": 'static/test_images/5.jpg', "content": "Such pretty colors!" },
        { "img": 'static/test_images/6.jpg', "content": "Eine URL: https://google.com" },
        { "img": 'static/test_images/7.jpg', "content": "🥐" },
    ]
    passed_cnt = 0
    failed_cnt = 0
    total_time = 0

    for test_image in test_images:
        print("="*60)
        print("Test image: %s" % test_image["img"])
        print("-"*60)
        str_data, version, read_time, raw_binary_string, output_image, binary_img = FGCReader.read_image(test_image["img"])
        # show_image("Output of: " + test_image["img"], output_image)
        # show_image("Binary of: " +  test_image["img"], binary_img)

        print("-"*60)
        print("RAW Read:   %s" % raw_binary_string)
        print("Version:    %s" % str(version))
        print("Read Time:  %.3fs" % read_time)
        print("Test data:  %s" % str(list(test_image["content"])))
        print("Read data:  %s" % str(list(str_data)))
        print("Result:     ", end="")
        if str_data == test_image["content"]:
            print("PASSED.")
            passed_cnt += 1
        else:
            print("FAILED.")
            failed_cnt += 1
        
        total_time += read_time 
    
    print("="*60)
    print("TOTAL PASSED:  %i" % passed_cnt)
    print("TOTAL FAILED:  %i" % failed_cnt)
    print("AVERAGE TIME:  %.3fs" % (total_time / len(test_images)))

    # plt.show()

if __name__ == '__main__':
    main()